var _ = require('underscore'),
    Buffer = require('buffer').Buffer,
    deferred = require('deferred'),
    redis = require('redis'),
    ns = require('./ns'),
    redisPayload = require('./redis_payload'),
    pluginsMap = {},
    IOCTL_REQ_NUM = 0,
    IOCTL_REQ_NUM_COUNT = 0,
    IOCTL_REQ_NUM_MAX = 4299999999999999,
    MAX_NUM_PLUGINS = 16,
    SPUTNIK_REQ_NUM_PREFIX = "42",
    ioctlAddrNum = 150,
    ioctl,


    IOCTL_REDIS_ADDRESS = 0xff,
    BONDINF_CMDID_ECHO = 0,
    IOCTL_CMD = BONDINF_CMDID_API_REQUEST = 1,
    IOCTL_RESTART_CODEC = BONDINF_CMDID_RESTART_VIDEO = 2,
    BONDINF_CMDID_SNAPSHOT = 3,
    BONDINF_CMDID_CONTROL = 4,
    BONDINF_CMDID_CUBE_ID = 5,
    BONDINF_CMDID_PUBLISH = 6,
    BONDINF_CMDID_SYSTEM = 7,
    BONDINF_CMDID_REPLY = 8,
    BONDINF_CMDID_CODEC = 9,
    BONDINF_CMDID_FILE_ACK = 10,
    BONDINF_CMDID_UPLOADER = 11,
    BONDINF_CMDID_STORAGE = 12,
    BONDINF_CMDID_CONNECTION = 13,

/* --- Flags --- */

    IOCTL_LOW_REDUNDANCY = 0x1000,
    IOCTL_BINARY_FORMAT = 0x2000,
    IOCTL_REPLY_REQUESTED = 0x4000,
    IOCTL_LOW_PRIORITY = 0x8000,

/* --- RTMP plugin --- */

    IOCTL_REDIS_RTMPWRITER_CONNECT = (0x0003 | IOCTL_BINARY_FORMAT),
    IOCTL_REDIS_RTMPWRITER_DISCONNECT = (0x0004 | IOCTL_BINARY_FORMAT),
    IOCTL_REDIS_RTMPWRITER_GET_STATE = (0x0005 | IOCTL_BINARY_FORMAT | IOCTL_REPLY_REQUESTED),

/* --- Other ioctl commands --- */

    IOCTL_BOND_GET_STATS = (0x0345 | IOCTL_BINARY_FORMAT /*| IOCTL_REPLY_REQUESTED*/),
    IOCTL_PLUGIN_GET_CURRENT_FRAME_RELEASE_TIME = (0x0347 | IOCTL_BINARY_FORMAT /*| IOCTL_REPLY_REQUESTED*/),
    IOCTL_BOND_SET_BITRATE = (0x0343 | IOCTL_BINARY_FORMAT),
    IOCTL_BOND_SET_BITRATE_FUNC = (0x0346 | IOCTL_BINARY_FORMAT),
    IOCTL_REGISTER_CALLBACK_FUNC = (0x0342 | IOCTL_BINARY_FORMAT),
    IOCTL_MODEM_INFO = (0x0348 | IOCTL_BINARY_FORMAT),
    IOCTL_REDIS_MESSAGE = (0x0349 | IOCTL_BINARY_FORMAT),
    IOCTL_SET_BOND_PARAMETER = (0x034a | IOCTL_BINARY_FORMAT),
    IOCTL_SET_TS_SERVER_PORT = (0x034b | IOCTL_BINARY_FORMAT),
    IOCTL_SEND_CC = (0x034c | IOCTL_BINARY_FORMAT),
    IOCTL_ENABLE_STREAMING = (0x034d | IOCTL_BINARY_FORMAT),

/* --- Recording on-the-fly --- */

    IOCTL_SET_FILE_RECORDING_PARAM = (0x034e | IOCTL_BINARY_FORMAT),
    IOCTL_GET_FILE_RECORDING_PARAM = (0x034f | IOCTL_BINARY_FORMAT | IOCTL_REPLY_REQUESTED),

/* --- Plugin loading --- */

    IOCTL_GET_RUNNING_PLUGINS_LIST = (0x0350 | IOCTL_BINARY_FORMAT | IOCTL_REPLY_REQUESTED),
    IOCTL_GET_MAIN_PLUGIN_INIT_PARAMS = (0x0351 | IOCTL_BINARY_FORMAT | IOCTL_REPLY_REQUESTED),
    IOCTL_GET_ADDIT_PLUGIN_INIT_PARAMS = (0x0352 | IOCTL_BINARY_FORMAT | IOCTL_REPLY_REQUESTED),
    IOCTL_LOAD_PLUGIN = (0x0354 | IOCTL_BINARY_FORMAT),
    IOCTL_UNLOAD_PLUGIN = (0x0353 | IOCTL_BINARY_FORMAT),
    IOCTL_ADD_PLUGIN_INIT_PARAMS = (0x0355 | IOCTL_BINARY_FORMAT),
    IOCTL_REMOVE_PLUGIN_INIT_PARAMS = (0x0356 | IOCTL_BINARY_FORMAT),

/* --- Stream snapshots --- */

    IOCTL_GET_SNAPSHOT = (0x0359 | IOCTL_BINARY_FORMAT),


/* --- File Uploader Path -- */
    IOCTL_SET_FILEUPLOADER_PATH  = (0x035b | IOCTL_BINARY_FORMAT),

/* --- Cube plugin IDs --- */

    PLUGIN_BENDER_ID = 1,
    PLUGIN_BONDCMD_ID = 2,
    PLUGIN_BOND_ID = 4,
    PLUGIN_BOND_RTSP_ID = 6,
    PLUGIN_BOND_TS_ID = 7,
    PLUGIN_BOND_RTMP_ID = 8,
    PLUGIN_BOND_LS_ID = 9,
    PLUGIN_BOND_IFB_ID = 10,
    PLUGIN_TS_SERVER_ID = 11;





/**
 Creates a string used to publish IOCTL request to Sputnik
 @function formatRequest
 @param {String} ioctlId Unique IOCTL request id (increment)
 @param {Number} ioctlCommand IOCTL number to call
 @param {Number} attr One byte of attributes (0-low priority; 1-high priority) and 0x02 (0-send with redundancy 1; 1-send with full redundancy)
 @param {String | Object} data Object or JSON with data
 @return {String} Returns complete IOCTL request string which will be published to Redis
 */

function formatRequest(ioctlId, ioctlCommand, attr, data) {

    /*
     Current format of the IOCTL requests published to a channel is following:
     <message_type>:<req_num>:<ioctl_addr>:<ioctl_cmd>:<attr>:<data>

     <message_type> should be 'req' for IOCTL calls;
     <req_num> is a unique request serial number;
     <ioctl_addr> is IOCTL ID (could be ID of bondwriter2 plugin running under mobiapp; ID of bondwriter2 plugin running under daemon; ID of bond_client; or ID of Sputnik)
     <ioctl_cmd> is IOCTL number to call;
     <attr> is a one byte of attributes; currently: 0x01 (0-low priority; 1-high priority) and 0x02 (0-send with redundancy 1; 1-send with full redundancy). Recommended value for short and not too frequent tdcontrol commands is 3.
     <data> is the payload data to send.
     */

    var payload = [],
        data = (_.isObject(data)) ? JSON.stringify(data) : data;

    payload.push('req');
    payload.push(IOCTL_REQ_NUM);
    payload.push(ioctlId);
    payload.push(ioctlCommand);
    payload.push(attr);
    payload.push(data);

    return payload.join(':');
}

/**
 Creates a string to match Redis messages to
 @function formatResponse
 @return {String} Returns a string to use in matching replies from Redis
 */

function formatResponse() {
    var payload = [];

    payload.push('reply');
    payload.push(IOCTL_REQ_NUM);

    return payload.join(':');
}


var openRequests = {};
var cleanUpOpenRequests = function(){
    _.each(openRequests, function(el, k){
        if( el.resolved){ delete openRequests[k];}})
}


var generateReqNum = function(){
    IOCTL_REQ_NUM_COUNT++;
    var ioctlReqNum = parseInt( SPUTNIK_REQ_NUM_PREFIX+IOCTL_REQ_NUM_COUNT );
    if( ioctlReqNum >= IOCTL_REQ_NUM_MAX ){
        IOCTL_REQ_NUM_COUNT = 0;
        ioctlReqNum = parseInt( SPUTNIK_REQ_NUM_PREFIX+IOCTL_REQ_NUM_COUNT );
    }

    return ioctlReqNum;
}
/**
 Publishes IOCTL command to Redis and waits for response for this command from Redis
 @function ioctlCommand
 @param {String} channel Redis channel for Redis pub/sub: 'sess:<sid>:<interface E|D|C|I>'
 @param {String} ioctlId Unique IOCTL request id (increment)
 @param {Number} command IOCTL number to call
 @param {Object | String} data Payload for command
 @param {Object} def Deferred which needs to be resolved when reply to IOCTL command is received
 @param {Object} [options] Additional options for the command, e.g. priority and redundancy flags
 @return {Number} Returns used ioctlReqNum value
 */

function ioctlCommand(channel, ioctlId, command, data, def, options) {
    options = options || {};

    IOCTL_REQ_NUM = generateReqNum();

    //some requests, do not have a response (eg. IOCTL_ADD_PLUGIN_INIT_PARAMS )
    //so the passed in deferred is null
    if( def != null ){
        openRequests[IOCTL_REQ_NUM] = def;
    }

    var redisPub = redis.createClient(),
        pubDef   = deferred(),
        redisSub = redis.createClient(),
        subDef   = deferred(),
        priority = (options.priority) ? 1 : 0,
        redundancy = (options.redundancy) ? 2 : 0,
        reqStr = formatRequest(ioctlId, command, priority | redundancy, data),
        resStr = formatResponse(),
        waiting;


    redisSub.on('ready', function(){subDef.resolve(1);});
    redisPub.on('ready', function(){pubDef.resolve(1);});

    deferred(pubDef.promise, subDef.promise)(function(){

        redisSub.subscribe(channel);

        /* --- Listen to reply from Redis --- */
        redisSub.on('message', function (channel, data) {
            var isReply = data.match(/reply:([0-9]+):/);
            //if (data.indexOf(resStr) === 0) {
            var reqNum = (isReply)? isReply[1]:null;
            if (reqNum && openRequests[ reqNum ] ) {
                clearTimeout(waiting);
                waiting = null;
                var payload = redisPayload(data);
                if( reqNum in openRequests ){
                    var def = openRequests[reqNum];
                    if (payload.result && payload.result === 'error') {
                        def.resolve(new Error(payload.error));
                    } else {
                        def.resolve(payload);
                    }
                    delete openRequests[reqNum];
                }
                cleanUpOpenRequests();
                redisPub.quit();
                redisSub.removeAllListeners();
                redisSub.quit();
            }
        });

        /* --- Publish to Redis --- */
        redisPub.publish(channel, reqStr);
        console.log('ioctlCommand():', channel, reqStr);

        /* --- End listening socket --- */
        if( def ){
            waiting = setTimeout(function () {
                redisPub.quit();
                redisSub.removeAllListeners();
                redisSub.quit();
                def.resolve(new Error('ioctl end socket timeout on ' + channel + ' ' + resStr));
                waiting = null;
            }, 120000);
        }
    });


    return IOCTL_REQ_NUM;
}

/**
 Publishes IOCTL command to Redis and waits for binary response for this command from Redis (response may come in multiple parts)
 @function ioctlCommandBinary
 @param {String} channel Redis channel for Redis pub/sub: 'sess:<sid>:<interface E|D|C|I>'
 @param {String} ioctlId Unique IOCTL request id (increment)
 @param {Number} command IOCTL number to call
 @param {Object | String} data Payload for command
 @param {Object} def Deferred which needs to be resolved when reply to IOCTL command is received
 @param {Object} [options] Additional options for the command, e.g. priority and redundancy flags
 @return {Number} Returns used ioctlReqNum value
 */

function ioctlCommandBinary(channel, ioctlId, command, data, def, options) {
    options = options || {};
    IOCTL_REQ_NUM = generateReqNum();

    var redisPub = redis.createClient(),
        redisSub = redis.createClient(null, null, { return_buffers: true }),
        priority = (options.priority) ? 1 : 0,
        redundancy = (options.redundancy) ? 2 : 0,
        reqStr = formatRequest(ioctlId, command, priority | redundancy, data),
        resStr = formatResponse(),
        resChunks = [],
        resMsg = 0,
        waiting;

    /* --- Listen to reply from Redis --- */

    redisSub.subscribe(channel);

    redisSub.on('message', function (channel, data) {

        // data is a Buffer object

        if (data.toString().indexOf(resStr) === 0) {

            waiting = null;
            /*
             Format of the replies: binary data formatted as following:

             byte 0: type of the snapshot (currently 0 which is h264)
             byte 1: total number of frames that will be sent (currently 1)
             byte 2: current frame that is being sent (starts from 0)
             byte 3: how many parts of this frame will be sent
             byte 4: current part number that is being sent (starts from 0)
             byte 5: start of binary frame data
             */

            var bin = data.slice(resStr.length + 1),
                resParts = bin[3],
                resCur = bin[4],
                resData = bin.slice(5);

            resMsg++;
            resChunks[resCur] = resData;
            data = null;
            if (resMsg == resParts) {
                clearTimeout(waiting);

                var img = Buffer.concat(resChunks);
                def.resolve(img);
                redisPub.quit();
                redisSub.removeAllListeners();
                redisSub.quit();
                resChunks = null;
            }
        }
    });

    /* --- Publish to Redis --- */

    redisPub.publish(channel, reqStr);

//    console.log('iotclCommandBinary():', channel, reqStr);

    /* --- End listening socket --- */

    waiting = setTimeout(function () {
        redisPub.quit();
        redisSub.removeAllListeners();
        redisSub.quit();
        resChunks = null;
        def.resolve(new Error('ioctl end socket timeout on ' + channel + ' ' + resStr));
        waiting = null;
    }, 10000);

    return IOCTL_REQ_NUM;
}

/**
 Adds new plugin to plugins map
 @function addPluginToMap
 @param {String} pluginName Name of the plugin, e.g. 'udpts'
 @param {Number} ioctlId Number of ioctl request
 @param {String} sid Serial number of device for which this plugin is added
 @param {Object} [data] Optional plugin data
 */

function addPluginToMap(pluginName, pluginId, sid, data, status) {
     pluginsMap[pluginName] = pluginsMap[pluginName] || {};
     pluginsMap[pluginName][sid] = pluginsMap[pluginName][sid] || [];
     var plugin = {
        id: pluginId,
        data: data || null
    };
    if( status ){plugin.status = status; }
    //prevent duplicates
    var existing = _.findWhere(pluginsMap[pluginName][sid], {id:pluginId});
    if( !(existing) ){
        pluginsMap[pluginName][sid].push(plugin);
    }
    else{
        //update params
        existing.data = plugin.data;
    }
    //console.log('addPluginToMap()', pluginsMap);
}

function getPlugin(pluginName, sid) {
    if (pluginsMap[pluginName] && pluginsMap[pluginName][sid]) {
        return pluginsMap[pluginName][sid];
    } else {
        return undefined;
    }
}

function removePlugin(pluginId) {
    _.each(pluginsMap, function (pluginByName) {
        _.each(pluginByName, function (pluginBySid) {
            _.each(pluginBySid, function (plugin, index) {
                if (plugin.id == pluginId) {
                    pluginBySid = pluginBySid.splice(index, 1);
                    console.log('removePlugin() done', index);
                }
            });
        });
    });
}


function updatePluginMapStatus( runningPluginIds, sid ){
    _.each( pluginsMap, function (pluginByName) {
            if (sid in pluginByName) {
                var pluginBySid = pluginByName[sid];
                _.each(pluginBySid, function(plugin){
                    if( _.indexOf(runningPluginIds, plugin.id) != -1 ){
                        plugin.status = 'running';
                        //console.log('plugin status is running!', plugin)
                    }
                    else{
                        plugin.status = 'inactive';
                        //console.log('plugin status is INACTIVE!', plugin)
                    }
                })

            }

    })
}

/**
 Takes a session and returns ICOTL API
 @function ioctl
 @param {Object} session Device session object { sid, iface, ctlids, channel }
 */

ioctl = module.exports = function (session) {
    return {

        /**
         Gets type of device (encoder, decoder, vidiu, bond etc)
         @method getDeviceType
         @return {Function} Returns a promise which will be resolved with type of device
         */

        getDeviceType: function () {
            var def = deferred(),
                _def = deferred(),
                data = {
                    command: 'get',
                    q: 'System.Info.Product'
                };

            ioctlCommand(session.ch, session.ctlids.aux, IOCTL_CMD, data, _def);

            _def.promise(function (data) {
                var
                    type = 'unknown',
                    prodName = data.System.Info.Product.productname,
                    matchDec = prodName.match(/cube-[34][0-9][0-9]|clip\srx/gi),
                    matchEnc = prodName.match(/cube(-[1235][0-9][0-9]|\spro)|clip\stx|bond\sii\s(sdi|hdmi)|brik-[1][0-9][0-9]|chimera|slice/gi),
                    matchVidiu = prodName.match(/vidiu/gi),
                    matchBond = prodName.match(/bond\sii\scoproc/gi);

                if (matchDec) {
                    type = 'decoder';
                } else if (matchEnc) {
                    type = 'encoder';
                } else if (matchVidiu) {
                    type = 'vidiu';
                } else if (matchBond) {
                    type = 'bond';
                }

                def.resolve(type);
            }, function (err) {
                def.resolve(new Error('Fail to get device type for ' + session.ch, err.stack || err.toString()));
            });

            return def.promise;
        },

        /**
         Gets device settings using optional query
         @method getSettings
         @param {String} [query] Optional comma separated string with queries, e.g. 'Services.Zeroconf.friendlyname,...'
         @return {Function} Returns a promise which will be resolved with settings for the query
         */

        getSettings: function (query) {
            query = query || '';

            var def = deferred(),
                values = {
                    command: 'get',
                    q: query
                };

            ioctlCommand(session.ch, session.ctlids.aux, IOCTL_CMD, values, def);
            return def.promise;
        },

        /**
         Gets device settings using optional query
         @method getSettings
         @param {String} [query] Optional comma separated string with queries, e.g. 'Services.Zeroconf.friendlyname,...'
         @return {Function} Returns a promise which will be resolved with settings for the query
         */

        getSettingsIntrospect: function (query) {
            query = query || '';

            var def = deferred(),
                values = {
                    command: 'geti',
                    q: query
                };

            ioctlCommand(session.ch, session.ctlids.aux, IOCTL_CMD, values, def);
            return def.promise;
        },
        /**
         Sets device settings
         @method setSettings
         @param {String} data Object (namespaced) with settings to be applied
         @return {Function} Returns a promise which will be resolved with updated settings
         */

        setSettings: function (data) {

            console.log('setSettings()', data);

            var def = deferred();

            var values = _.extend({}, ns.stringify(data), {
                command: 'update'
            });

            ioctlCommand(session.ch, session.ctlids.aux, IOCTL_CMD, values, def);
            return def.promise;
        },

        /**
         Gets devices
         @param {String} [query] Optional comma separated string with queries, e.g. 'Services.Zeroconf.friendlyname,...'
         @return {Function} Returns a promise which will be resolved with settings for the query
         */

        getDevices: function () {

            var def = deferred(),
                values = {
                    command: 'devices'
                };

            ioctlCommand(session.ch, session.ctlids.aux, BONDINF_CMDID_UPLOADER, values, def);
            return def.promise;
        },

        /**
         Gets files (recordings) using optional query
         @param {String} [device] Required comma separated string with queries, e.g. 'Services.Zeroconf.friendlyname,...'
         @return {Function} Returns a promise which will be resolved with settings for the query
         */

        getFiles: function (device, dir) {
            var def = deferred(),
                values = {
                    command: 'files',
                    device: device,
                    dir: dir
                };

            ioctlCommand(session.ch, session.ctlids.aux, BONDINF_CMDID_UPLOADER, values, def);
            return def.promise;
        },

        /**
         get the list of files that have been set for transfer
         @param {String} [query] Optional comma separated string with queries, e.g. 'Services.Zeroconf.friendlyname,...'
         @return {Function} Returns a promise which will be resolved with settings for the query
         */

        getFileTransferQ: function () {
            var def = deferred(),
                values = {
                    command: 'list'
                };

            ioctlCommand(session.ch, session.ctlids.aux, BONDINF_CMDID_UPLOADER, values, def);
            return def.promise;
        },

        /**
         Add a file to the queue of files to be transferred from the remote device
         @param {String} [query] Optional comma separated string with queries, e.g. 'Services.Zeroconf.friendlyname,...'
         @return {Function} Returns a promise which will be resolved with settings for the query
         */

        addFileTransferQ: function (device, file) {
            var def = deferred(),
                values = {
                    command: 'add',
                    device:device,
                    file:file
                };

            ioctlCommand(session.ch, session.ctlids.aux, BONDINF_CMDID_UPLOADER, values, def);
            return def.promise;
        },


        /**
         remove a file from the transfer queue
         @method removeFileTransfer
         @param {String} [device] the device
         @param {String} [file] the path of the file relative to the root
         @return {Function} Returns a promise which will be resolved with settings for the query
         */

        removeFileTransferQ: function (device, file) {
            var def = deferred(),
                values = {
                    command: 'remove',
                    device:device,
                    file:file
                };

            ioctlCommand(session.ch, session.ctlids.aux, BONDINF_CMDID_UPLOADER, values, def);
            return def.promise;
        },


        /**
         Add a File to be transferred from the remote device
         @method removeFileTransfer
         @param {String} [query] Optional comma separated string with queries, e.g. 'Services.Zeroconf.friendlyname,...'
         @return {Function} Returns a promise which will be resolved with settings for the query
         */

        getFileTransferQStatus: function (device, file) {
            var def = deferred(),
                values = {
                    command: 'add',
                    device:device,
                    file:file
                };

            ioctlCommand(session.ch, session.ctlids.aux, BONDINF_CMDID_UPLOADER, values, def);
            return def.promise;
        },

/*
 * <plugin=snap,id=36>
 * <plugin=snap,id=37>
 * <id=152,plugin=snap>
 * <id=153,plugin=snap>
 * <id=154,plugin=snap>
 * <id=155,plugin=rtp,ses=15001429,prt=554,vid=h264,aud=aac,mip=NULL,htp=0>
 *
 * returns {<pluginName>:[[<plugin instance params>], [<plugin instance params>]]}
 */
        _parsePluginParams: function(params){
            var res = {};
            var plugins = params.split(/<(.*?)>/g);
            _.map( plugins, function(plugin){
                var pluginName = plugin.match(/(?:\bplugin=)([^,|>]+)/);
                if( pluginName ){
                    pluginName = pluginName[1];
                    res[pluginName] = res[pluginName] || [];
                    res[pluginName].push(plugin.split(','));
                }
            })
            return res;
        },

        _pluginAlreadyLoaded: function( pluginName, pluginParams ){
            var existingPlugins = getPlugin(pluginName, session.sid);
            if( _.isArray(existingPlugins) && existingPlugins.length >= 1 ){
                var matchingPluginId = false;
                _.each(existingPlugins, function(plugin){
                    //get plugins with id's above 150 since we are allowing
                    //core and sputnik to load duplicate plugins
                    //and are separating them by address range
                    //instead of sharing
                    // "inactive" plugins are ok since they will be re-activated
                    if( parseInt(plugin.id) >= 150 && plugin.status == 'running' ){
                        var params = plugin.data;
                        //remove unique fields id & plugin (pluginName)
                        var cleanParams = _.chain(params).uniq().reject(function(p){ return p.match(/^id=.*|plugin=.*$/); }).value();
                        pluginParams = _.uniq(pluginParams);
                        var isUnique = (cleanParams.length != pluginParams.length)
                                || (_.difference(cleanParams, pluginParams).length != 0);

                        if( !(isUnique) ){
                            matchingPluginId = plugin.id;
                        }
                        return isUnique;
                    }
                });
                return matchingPluginId;
            }
            return false;
        },

        _pluginAlreadyLoadedParams: function( existingPluginParams, pluginName, pluginParams ){
            var existingPlugins = this._parsePluginParams(existingPluginParams);

            if( pluginName in existingPlugins ){
                var samePluginType = existingPlugins[pluginName];
                var match = false;
                //is every plugin different from plugin to load
                var everyPluginUnique = _.every(samePluginType, function(params){
                    //remove unique fields id & plugin (pluginName)
                    var cleanParams = _.chain(params).uniq().reject(function(p){ return p.match(/^id=.*|plugin=.*$/); }).value();
                    pluginParams = _.uniq(pluginParams);
                    var isUnique = (cleanParams.length != pluginParams.length)
                            || (_.difference(cleanParams, pluginParams).length != 0);

                    if( !(isUnique) ){
                        match = _.find(params, function(v){return v.match(/^(?:id=)(\d+)/);}).split('=')[1];
                    }
                    return isUnique;
                });
                return match;
            }
            return false;
        },
        /**
         Sets device settings
         @method addPluginParams
         @param {String} pluginName Name of the plugins, e.g. 'udpts', 'rtmp', etc.
         @param {Array} params Array with params for the plugin
         @return {Function} Returns a promise resolved with new plugin id
                /*
                 To add initialization string to additional initialization strings list:
                 Request destination: Sputnik;
                 command: IOCTL_ADD_PLUGIN_INIT_PARAMS;
                 payload: JSON statement like {"cube_id":<sid>, "init_params":"<plugin=udpts,id=34,…><…>"}
                 If cube_id is 0, the setting is valid for all Cubes which cube_ids are not listed individually
                 (see description of the configuration file in the e-mail below). If an entry with the same cube_id
                 already exists in the list, it is removed first. Any parameter in initialization string may contain
                 combination %!CUBE_ID, that gets replaced by Cube serial number (cube_id) before it is passed to plugins.
                 No spaces are allowed except inside initialization string. Plugins that are already loaded continue
                 to run even if their initialization string is updated.

         */

        addPluginParams: function (pluginName, params, force) {
            console.log('addPluginParams()', pluginName, params);
            var def = deferred();
            var self = this;
            this.listRunningPluginParams()(function(curPlugins){
                var paramsStr = '';
				var existingSessionPlugins = _.find(curPlugins.addit_plugin_params, function(p){return p.cube_id == session.sid; });
				if( existingSessionPlugins ){
                 	paramsStr = existingSessionPlugins.init_params;
                    var existingPlugin = self._pluginAlreadyLoaded(pluginName, params);
                    if( existingPlugin ){

                        console.log("Note: There is already an existing '%s' plugin with id:%s", pluginName, existingPlugin);
                        if( !(force) ){
                            def.resolve(new Error("Note: There is already an existing '"+pluginName+"' plugin with id:"+existingPlugin) );
                            return def.promise;
                        }
                        console.log("Note: forcing the loading of a duplicate '%s' plugin", pluginName);

                        var usedIoctlAddrs = _.map(paramsStr.match(/id=\d+/g), function(v){return parseInt(v.split('=')[1])});
                        if( usedIoctlAddrs.length > 0 ){
                            //start from the max of the used up ioctl addrs
                            ioctlAddrNum = _.max(usedIoctlAddrs);
                        }
                    }
				}

                ioctlAddrNum++;
                ( ioctlAddrNum < 150 || ioctlAddrNum > 200 ) ? ioctlAddrNum = 150 : ioctlAddrNum;

                params.unshift('plugin=' + pluginName);
                params.unshift('id=' + ioctlAddrNum);

                paramsStr += '<' + params.join(',') + '>';

                var data = {
                        cube_id: session.sid,
                        init_params: paramsStr
                    },
                    options = {
                        priority: 1,
                        redundancy: 1
                    };

                self.modifyPluginParams(data);
                addPluginToMap(pluginName, ioctlAddrNum, session.sid, params);
                def.resolve(ioctlAddrNum);
            })
            return def.promise;
        },

        /*
         * This ioctl request does not have a corresponding response,
         * so sending null for the deferred object in ioctlCommand
         */
        modifyPluginParams: function(data){
              var options = {
                            priority: 1,
                            redundancy: 1
                 };
             ioctlCommand(session.ch, session.ctlids.sputnik, IOCTL_ADD_PLUGIN_INIT_PARAMS, data, null, options);
        },

        /**
         Sets device settings
         @method addPluginParams
         @param {Array} params Array with params for the plugin
         @return {Function} Returns a resolved promise
         */

        removePluginParams: function () {

            /*
             To remove initialization string from additional initialization strings list:
             Request destination: Sputnik;
             command: IOCTL_REMOVE_PLUGIN_INIT_PARAMS;
             payload: JSON statement {"cube_id": <sid>}
             If an entry with the same cube_id exists in the list, it is removed. Cube_id can be 0.
             Plugins that are already loaded continue to run even if their initialization string is removed.
             */

            console.log('removePluginParams()');

            var def = deferred(),
                data = {
                    cube_id: session.sid
                };

            ioctlCommand(session.ch, session.ctlids.sputnik, IOCTL_REMOVE_PLUGIN_INIT_PARAMS, data, def);
            def.resolve(1);
            return def.promise;
        },

        /**
         Loads currently running plugin
         @method loadPlugin
         @param {String} ioctlId IOCTL ID of the plugin to load
         @return {Function} Returns a resolved promise
         */

        loadPlugin: function (ioctlId) {

            /*
             To load currently running plugin:
             Request destination: Sputnik;
             command: IOCTL_LOAD_PLUGIN;
             payload: one or several JSON statements e.g. "ioctl_id":123
             Loads plugin with a given IOCTL ID. Initialization string for this IOCTL ID must be present either
             in main initialization string (the one that is received from Cube), or in one of additional
             initialization strings (with the matching Cube ID). If a plugin with the same IOCTL ID is already
             loaded, it gets unloaded first.
             */

            console.log('loadPlugin()', ioctlId);

            var def = deferred(),
                data = {
                    ioctl_id: ioctlId
                },
                options = {
                    priority: 1,
                    redundancy: 1
                };

            ioctlCommand(session.ch, session.ctlids.sputnik, IOCTL_LOAD_PLUGIN, data, def, options);
            def.resolve(1);
            return def.promise;
        },

        /**
         Unloads currently running plugin
         @method loadPlugin
         @param {String} ioctlId IOCTL ID of the plugin to unload
         @return {Function} Returns a resolved promise
         */

        unloadPlugin: function (ioctlId) {
            var self = this;
            var def = deferred();
            /*
             To unload currently running plugin must first update the ioctl plugin init params to remove init params entry.
             Request destination: Sputnik;
             command: IOCTL_UNLOAD_PLUGIN;
             payload: one or several JSON statements e.g. "ioctl_id":123
             Unloads plugin with a given IOCTL ID.
             */
            this.listPluginParams()(function(curPlugins){
                var existing = _.find(curPlugins.addit_plugin_params, function(p){return p.cube_id == session.sid; });
                if( existing ){
                    var regex = new RegExp('<[^>]*id='+ioctlId+'[^>]*>', 'g');
                    var removed = existing.init_params.replace(regex,'');
                    self.modifyPluginParams(removed);
                }

                console.log('unloadPlugin()', ioctlId);

                 var data = {
                        ioctl_id: ioctlId
                    },
                    options = {
                        priority: 1,
                        redundancy: 1
                    };

                ioctlCommand(session.ch, session.ctlids.sputnik, IOCTL_UNLOAD_PLUGIN, data, def, options);
                def.resolve(1);
                removePlugin(ioctlId);
            });
            return def.promise;
        },

        /**
         Unloads all currently loaded plugins for this session
         @method unloadAllPlugins
         */

        unloadAllPlugins: function () {
            _.each(pluginsMap, function (pluginByName) {
                if (pluginByName[session.sid]) {
                    _.each(pluginByName[session.sid], function (plugin) {
                        var def = deferred(),
                            data = {
                                ioctl_id: plugin.id
                            },
                            options = {
                                priority: 1,
                                redundancy: 1
                            };

                        ioctlCommand(session.ch, session.ctlids.sputnik, IOCTL_UNLOAD_PLUGIN, data, def, options);
                        removePlugin(ioctlId);
                        def.resolve(1);
                    });
                }
            });
        },

        unloadPluginsMap: function(){
            if( _.isEmpty(pluginsMap) ){ return true; }
            _.each(pluginsMap, function (pluginByName) {
                if (pluginByName[session.sid]) {
                    _.each(pluginByName[session.sid], function (plugin) {
                        removePlugin(plugin.id);
                    });
                }
            });
            return true;
        },
        /**
         Lists IOCTL IDs of currently running plugins
         @method loadPlugin
         @param {String} ioctlId IOCTL ID of the plugin to unload
         @return {Function} Returns a resolved promise
         */

        listPlugins: function () {
            /*
             To get a list of IOCTL IDs of the plugins loaded in the current session:
             Send a request via Redis with destination set to Sputnik;
             command: IOCTL_GET_RUNNING_PLUGINS_LIST;
             payload: none
             Reply will be in JSON format, e.g.:
             reply:125:{"plugin_ids":[7,11]}, where 7 and 11 are IDs of the loaded (and currently running) plugins.
             */

            console.log('listPlugins()', pluginsMap);
            return pluginsMap;
        },


/*
             To get a list of IOCTL IDs of the plugins loaded on sputnik (not just for this session):
             Send a request via Redis with destination set to Sputnik;
             command: IOCTL_GET_RUNNING_PLUGINS_LIST;
             payload: none
             Reply will be in JSON format, e.g.:
             reply:125:{"plugin_ids":[7,11]}, where 7 and 11 are IDs of the loaded (and currently running) plugins.
             */
        listRunningPluginIds: function(){

            var def = deferred(),
                data = {
                },
                options = {
                    priority: 1,
                    redundancy: 1
                };

            ioctlCommand(session.ch, session.ctlids.sputnik, IOCTL_GET_RUNNING_PLUGINS_LIST, data, def, options);
            def.promise(function(curRunningIds){updatePluginMapStatus(curRunningIds.plugin_ids, session.sid);});
            return def.promise;

        },

        /**
         Takes the list of plugin parameters and then filtering it to list only the ones that are actually currently running
         @method listRunningPluginParams
         @return {Function} Returns a resolved promise
         */

        listRunningPluginParams:function(){
            console.log("listRunningPluginParams()")
            var def = deferred();
            var self = this;
            deferred(this.listRunningPluginIds(),
            this.listMainPluginParams(),
                                 this.listPluginParams())(

                function(result){

                var curRunningIds = result[0].plugin_ids;
                var curMainPlugins = result[1].main_plugin_params;
                var curPluginParams = result[2].addit_plugin_params;

                var curSessionPlugin = _.find(curPluginParams, function(p){return p.cube_id == session.sid; });

                if( curSessionPlugin ){
                    var pluginParams = curSessionPlugin.init_params+curMainPlugins;
                    var plugins = pluginParams.split(/<(.*?)>/g);

                    _.each(plugins, function( plugin ){
                        if( plugin != "" ){
                            var idNum = parseInt(plugin.match(/(?:\bid=)([^,|>]+)/)[1]);
                            //if plugin in init_params but not running
                            if( _.indexOf(curRunningIds, idNum) == -1 ){
                                var regex = new RegExp('<[^<]*?id='+idNum+'[^>]*>', 'g');
                                curSessionPlugin.init_params = curSessionPlugin.init_params.replace(regex, '');
                                addPluginToMap(pluginName, idNum, session.sid, params, 'inactive');
                                //mark the plugin as inactive but could maybe remove the plugin
                                //removePlugin(idNum);
                            }
                            else{ //plugin is running and in init_params
                                var pluginName = plugin.match(/(?:\bplugin=)([^,|>]+)/)[1];
                                var params = plugin.split(',');
                                addPluginToMap(pluginName, idNum, session.sid, params, 'running');
                            }
                        }
                    });
                }

                def.resolve({addit_plugin_params:curPluginParams});
            }).done();
            return def.promise;

        },
        /**
         Lists IOCTL IDs of currently running plugins
         @method loadPlugin
         @param {String} ioctlId IOCTL ID of the plugin to unload
         @return {Function} Returns a resolved promise
         */

        listPluginParams: function () {

            /*
             To get a list of IOCTL ID plugin parameters
             Send a request via Redis with destination set to Sputnik;
             command: IOCTL_GET_ADDIT_PLUGIN_INIT_PARAMS;
             payload: none
             Reply will be in JSON format, e.g.:

             */

            console.log('listPluginParams()');
            var def = deferred(),
                data = {
                },
                options = {
                    priority: 1,
                    redundancy: 1
                };

            ioctlCommand(session.ch, session.ctlids.sputnik, IOCTL_GET_ADDIT_PLUGIN_INIT_PARAMS, data, def, options);

            return def.promise;
        },
        /**
         Lists IOCTL IDs of currently running plugins
         @method loadPlugin
         @param {String} ioctlId IOCTL ID of the plugin to unload
         @return {Function} Returns a resolved promise
         */

        listMainPluginParams: function () {

            console.log('listMainPluginParams()');
            var def = deferred(),
                data = {
                },
                options = {
                    priority: 1,
                    redundancy: 1
                };

            ioctlCommand(session.ch, session.ctlids.sputnik, IOCTL_GET_MAIN_PLUGIN_INIT_PARAMS, data, def, options);

            return def.promise;
        },




        /**
         Returns the LAST loaded ACTIVE plugin of the type 'pluginName'
         @method getActivePlugin
         @param {String} pluginName Name of plugin
         @return {Array} Array with plugins
         */

        getActivePlugin: function (pluginName) {
            var plugins = getPlugin(pluginName, session.sid);
            return _.find(plugins, function(p){ return ('status' in p) && (p.status == 'running') && (parseInt(p.id)>=150);}) || _.last(plugins);
        },

        /**
         Starts / stops streaming to CDN from Sputnik
         @method rtmpwriter
         @param {String} sessId Unique name of the session used by RTMP plugin
         */

        rtmpwriter: function (sessId) {

            /*
             Note that when dealing with plugins, 2nd argument of ioctlCommand() is not an IOCTL id for 'cube', 'aux', 'bond', 'sputnik';
             This is a plugin ID returned by addPluginParams()
             */

            var data = [
                    'ip=1.15940683.fme.ustream.tv',
                    'session=' + sessId,
                    'port=1935',
                    'file=ustreamVideo/15940683',
                    'user_agent=FMLE',
                    'copyright=Teradek',
                    'device=cube2',
                    'compat=0',
                    'strm=dNwWKPNSgycNvK6cAJbGwXVX2ugrscVV',
                    'pub.platform=RTMP',
                    'pub.username=',
                    'pub.password=',
                    'pub.auto_start=1'
                ],
                dataStr = data.join(','),

                runCommand = function (pluginId, command, def, params) {
                    var options = {
                        priority: 1,
                        redundancy: 1
                    };

                    var dataStr = '';
                    if (params) {
                        if( params.indexOf('session=' + sessId) == -1 ){
                            params.push('session=' + sessId);
                        }
                        dataStr = params.join(',');

                    }

                    ioctlCommand(session.ch, pluginId, command, dataStr, def, options);

                    if ((command | IOCTL_REPLY_REQUESTED) !== command) {
                        def.resolve(1);
                    }
                };

            return {

                /**
                 Sends "connect" command to this plugin
                 @method connect
                 @param pluginId
                 @return {Function} Returns resolved promise
                 */

                connect: function (pluginId, params) {
                    console.log("rtmpwriter.connect()", params)
                    var _def = deferred()
                    runCommand(pluginId, IOCTL_REDIS_RTMPWRITER_CONNECT, _def, params);
                    return _def.promise;
                },

                /**
                 Sends "disconnect" command to this plugin
                 @method disconnect
                 @param pluginId
                 @return {Function} Returns resolved promise
                 */

                disconnect: function (pluginId) {
                    console.log("rtmpwriter.disconnect()", pluginId)
                    var _def = deferred()
                    runCommand(pluginId, IOCTL_REDIS_RTMPWRITER_DISCONNECT, _def);
                    return _def.promise;
                },

                /**
                 Queries state of connection of RTMP plugin
                 @method getState
                 @param pluginId
                 @return {Function} Returns promise resolved with reply from IOCTL (0: disconnected, 1: connected, 2: connecting, 3: reconnecting)
                 */

                getState: function (pluginId, params) {
                    var _def = deferred()
                    runCommand(pluginId, IOCTL_REDIS_RTMPWRITER_GET_STATE, _def, params);
                    return _def.promise;
                }
            }
        },

        /**
         Queries state of connection of RTMP plugin
         @method getState
         @param options
          <type>      Snapshot type (h264,jpeg,ppm)
          <width>     JPEG Width
          <height>    JPEG Height
          <quality>   JPEG Quality
          <filename>  file to place snapshot

         @return {Function} Returns promise resolved with reply from IOCTL
         */
        takeSnapshot: function (options) {
            var def = deferred();
            if (pluginsMap && pluginsMap['snap'] && pluginsMap['snap'][session.sid]) {
                var activePlugin = this.getActivePlugin('snap');
                var pluginId = activePlugin.id;
                var dataStr = _.defaults(options, {type: "jpeg"});
                    dataStr = JSON.stringify(options);
                    options = {
                        priority: 1,
                        redundancy: 1
                    };
            ioctlCommandBinary(session.ch, pluginId, IOCTL_GET_SNAPSHOT, dataStr, def, options);

            } else {
                def.resolve(new Error('takeSnapshot() snap plugin not loaded'));
            }

            return def.promise;

        },

        setRecordParams: function (prm) {
            var def = deferred();
            if (pluginsMap && pluginsMap['udpts'] && pluginsMap['udpts'][session.sid]) {
                var pluginId = getPlugin('udpts', session.sid)[0].id,
                    dataStr = JSON.stringify(prm),
                    options = {
                        priority: 1,
                        redundancy: 1
                    };
                console.log('Setting record params for ' + session.sid + ':', dataStr);
                ioctlCommand(session.ch, pluginId, IOCTL_SET_FILE_RECORDING_PARAM, dataStr, def, options);
            } else {
                def.resolve(new Error('setRecorsdParams() udpts plugin not loaded'));
            }

            def.promise(function () {
                def.resolve.apply(def, arguments);
            }, function (err) {
                console.log(err.stack || err.toString());
            });
            return def.promise;
        },

        getRecordParams: function (prm) {
            var def = deferred();
            if (pluginsMap && pluginsMap['udpts'] && pluginsMap['udpts'][session.sid]) {
                var pluginId = getPlugin('udpts', session.sid)[0].id,
                    dataStr = JSON.stringify(prm),
                    options = {
                        priority: 1,
                        redundancy: 1
                    };
                console.log('Getting record params for ' + session.sid + ':', dataStr);
                ioctlCommand(session.ch, pluginId, IOCTL_GET_FILE_RECORDING_PARAM, dataStr, def, options);
            } else {
                def.resolve(new Error('getRecorsdParams() udpts plugin not loaded'));
            }

            def.promise(function () {
                def.resolve.apply(def, arguments);
            }, function (err) {
                console.log(err.stack || err.toString());
            });
            return def.promise;
        },

        restartCodec: function () {
            var def = deferred(),
                values = {};

            ioctlCommand(session.ch, session.ctlids.aux, IOCTL_RESTART_CODEC, values, def);
            def.resolve(1);
            return def.promise;
        },

        rebootDevice: function () {
            var def = deferred(),
                values = {
                    "command":"reboot",
                    "delay":"10"
                };

            ioctlCommand(session.ch, session.ctlids.aux, BONDINF_CMDID_SYSTEM, values, def);
            def.resolve(1);
            return def.promise;
        }

    }
}